@@ -129,10 +129,15 @@ class SLocation:
     ) -> None:
         if isinstance(locations, Mapping):
             frequencies, sigmas = list(locations.keys()), list(locations.values())
-        elif len(locations) % 2:
-            raise CustomValueError("locations must be an even number of items, being tuples.", self.__class__)
         else:
             locations = list[float](flatten(locations))  # type: ignore [arg-type]
+
+            if len(locations) % 2:
+                raise CustomValueError(
+                    "slocations must resolve to an even number of total items, pairing frequency and sigma respectively",
+                    self.__class__
+                )
+
             frequencies, sigmas = list(locations[0::2]), list(locations[1::2])
 
         frequencies = self.boundsCheck(frequencies, (0, 1), strict)
