@@ -7,15 +7,15 @@ from __future__ import annotations
 from dataclasses import dataclass
 from itertools import chain
 from math import ceil, exp
-from typing import TYPE_CHECKING, Any, Literal, Sequence, TypeVar, cast, overload
+from typing import TYPE_CHECKING, Any, Literal, NamedTuple, Sequence, TypeVar, cast, overload
 
 from _collections_abc import dict_items, dict_keys, dict_values
 from vstools import (
     MISSING, ColorRange, ConstantFormatVideoNode, CustomEnum, CustomIntEnum, CustomOverflowError, CustomStrEnum,
     CustomValueError, FieldBased, FieldBasedT, FuncExceptT, GenericVSFunction, InvalidColorFamilyError, KwargsNotNone,
-    KwargsT, MissingT, VSFunction, check_ref_clip, check_variable, clamp, core, depth, disallow_variable_format,
-    disallow_variable_resolution, fallback, inject_self, kwargs_fallback, normalize_planes, normalize_seq, scale_value,
-    vs
+    KwargsT, MissingT, PlanesT, VSFunction, check_ref_clip, check_variable, clamp, core, depth,
+    disallow_variable_format, disallow_variable_resolution, fallback, inject_self, kwargs_fallback, normalize_planes,
+    normalize_seq, scale_value, vs
 )
 
 from .prefilters import PelType, Prefilter, prefilter_to_full_range
@@ -25,7 +25,8 @@ __all__ = [
     'MVTools', 'MVToolsPlugin',
     'SADMode', 'SearchMode', 'MotionMode',
     'MVDirection', 'MotionVectors',
-    'MVToolsPresets'
+    'MVToolsPresets',
+    'SuperClips'
 ]
 
 
@@ -49,9 +50,6 @@ class MotionVectors:
     vmulti: vs.VideoNode
     """Super analyzed clip."""
 
-    super_render: vs.VideoNode
-    """Super clip used for analyzing."""
-
     kwargs: dict[str, Any]
 
     temporal_vectors: dict[MVDirection, dict[int, vs.VideoNode]]
@@ -108,7 +106,6 @@ class MotionVectors:
         """Deletes all :py:class:`vsdenoise.mvtools.MotionVectors` attributes."""
 
         del self.vmulti
-        del self.super_render
         self.kwargs.clear()
         self.temporal_vectors.clear()
         self._init_vects()
@@ -294,6 +291,8 @@ class MotionMode:
 
         Internally it is coefficient for SAD penalty of vector squared
         difference from predictors (neighbors), scaled by 256.
+
+        This is the "lambda" parameter of MVTools.
         """
 
         sad_limit: int
@@ -304,6 +303,8 @@ class MotionMode:
         is greater than the limit. It prevents bad predictors using but decreases the motion coherence.
 
         Values above 1000 (for block size=8) are recommended for true motion.
+
+        This is the "lsad" parameter of MVTools.
         """
 
         pnew: int
@@ -333,6 +334,10 @@ class MotionMode:
         Coherence is not used for global vector.
         """
 
+        def block_coherence(self, block_size: int) -> int:
+            """Method to calculate coherence (lambda) based on blocksize."""
+            return (self.coherence * block_size ** 2) // 64
+
     HIGH_SAD = Config(False, 0, 400, 0, 0, False)
     """Use to search motion vectors with best SAD."""
 
@@ -695,6 +700,13 @@ class MVToolsPresets:
     """CMDegrain from EoE."""
 
 
+class SuperClips(NamedTuple):
+    base: vs.VideoNode
+    render: vs.VideoNode
+    search: vs.VideoNode
+    recalculate: vs.VideoNode
+
+
 class MVTools:
     """MVTools wrapper for motion analysis / degrain / compensation"""
 
@@ -722,7 +734,7 @@ class MVTools:
         self, clip: vs.VideoNode,
         tr: int = 2, refine: int = 3, pel: int | None = None,
         planes: int | Sequence[int] | None = None,
-        range_in: ColorRange = ColorRange.LIMITED,
+        range_in: ColorRange | None = None,
         source_type: FieldBasedT | None = None,
         high_precision: bool = False,
         hpad: int | None = None, vpad: int | None = None,
@@ -734,15 +746,15 @@ class MVTools:
         analyze_args: dict[str, Any] | None = None,
         recalculate_args: dict[str, Any] | None = None,
         compensate_args: dict[str, Any] | None = None,
+        # super kwargs
+        range_conversion: float | None = None, sharp: int | None = None,
+        rfilter: int | None = None, prefilter: Prefilter | vs.VideoNode | None = None,
+        pel_type: PelType | tuple[PelType, PelType] | None = None,
         # analyze kwargs
         block_size: int | None = None, overlap: int | None = None,
-        thSAD: int | None = None, range_conversion: float | None = None,
-        search: SearchMode | SearchMode.Config | None = None,
-        sharp: int | None = None, rfilter: int | None = None,
+        thSAD: int | None = None, search: SearchMode | SearchMode.Config | None = None,
         sad_mode: SADMode | tuple[SADMode, SADMode] | None = None,
         motion: MotionMode.Config | None = None,
-        prefilter: Prefilter | vs.VideoNode | None = None,
-        pel_type: PelType | tuple[PelType, PelType] | None = None
     ) -> None:
         """
         MVTools is a wrapper around the Motion Vector Tools plugin for VapourSynth,
@@ -769,7 +781,8 @@ class MVTools:
         :param tr:                  Temporal radius of the processing.
         :param refine:              This represents the times the analyzed clip will be recalculated.\n
                                     With every recalculation step, the block size will be further refined.\n
-                                    i.e. `refine=4` it will analyze at `block_size=32`, then refine at 16, 8, 4.
+                                    i.e. `refine=3` it will analyze at `block_size=32`, then refine at 16, 8, 4.
+                                    Set `refine=0` to disable recalculation completely.
         :param pel:                 Pixel EnLargement value, a.k.a. subpixel accuracy of the motion estimation.\n
                                     Value can only be 1, 2 or 4.
                                      * 1 means a precision to the pixel.
@@ -849,7 +862,7 @@ class MVTools:
             raise CustomOverflowError(f'Refine > 6 is not supported! ({refine})', self.__class__)
 
         self.source_type = FieldBased.from_param(source_type, MVTools) or FieldBased.from_video(self.clip)
-        self.range_in = range_in
+        self.range_in = range_in or ColorRange.from_video(self.clip, False, self.__class__)
 
         self.pel = fallback(pel, 1 + int(not self.is_hd))
 
@@ -861,13 +874,6 @@ class MVTools:
 
         self.chroma = self.mv_plane != 0
 
-        if isinstance(vectors, MVTools):
-            self.vectors = vectors.vectors
-        elif isinstance(vectors, MotionVectors):
-            self.vectors = vectors
-        else:
-            self.vectors = MotionVectors()
-
         self.params_curve = params_curve
 
         self.super_args = fallback(super_args, dict[str, Any]())
@@ -893,23 +899,118 @@ class MVTools:
 
         self.mvtools = MVToolsPlugin.from_video(self.workclip)
 
-        self.analyze_func_kwargs = dict(
-            rfilter=rfilter, overlap=overlap, range_conversion=range_conversion, search=search, sharp=sharp,
-            block_size=block_size, sad_mode=sad_mode, motion=motion, prefilter=prefilter, pel_type=pel_type,
-            thSAD=thSAD
+        self.super_func_kwargs = dict(
+            rfilter=rfilter, range_conversion=range_conversion, sharp=sharp,
+            prefilter=prefilter, pel_type=pel_type
         )
 
-    def analyze(
-        self,
-        block_size: int | None = None, overlap: int | None = None,
-        thSAD: int | None = None, range_conversion: float | None = None,
-        search: SearchMode | SearchMode.Config | None = None,
-        sharp: int | None = None, rfilter: int | None = None,
-        sad_mode: SADMode | tuple[SADMode, SADMode] | None = None,
-        motion: MotionMode.Config | None = None,
-        prefilter: Prefilter | vs.VideoNode | None = None,
+        self.supers: SuperClips | None = None
+
+        self.analyze_func_kwargs = dict(
+            overlap=overlap, search=search, block_size=block_size, sad_mode=sad_mode,
+            motion=motion, thSAD=thSAD
+        )
+
+        if isinstance(vectors, MVTools):
+            self.vectors = vectors.vectors
+        elif isinstance(vectors, MotionVectors):
+            self.vectors = vectors
+        else:
+            self.vectors = MotionVectors()
+
+    def super(
+        self, range_conversion: float | None = None, sharp: int | None = None,
+        rfilter: int | None = None, prefilter: Prefilter | vs.VideoNode | None = None,
         pel_type: PelType | tuple[PelType, PelType] | None = None,
         *, ref: vs.VideoNode | None = None, inplace: bool = False
+    ) -> SuperClips:
+        """
+        Calculates Super clips for rendering, searching, and recalculating.
+
+        :param range_conversion:    If the input is limited, it will be converted to full range
+                                    to allow the motion analysis to use a wider array of information.\n
+                                    This is for deciding what range conversion method to use.
+                                     * >= 1.0 - Expansion with expr based on this coefficient.
+                                     * >  0.0 - Expansion with retinex.
+                                     * <= 0.0 - Simple conversion with resize plugin.
+        :param sharp:               Subpixel interpolation method for pel = 2 or 4. Possible values are 0, 1, 2.\n
+                                     * 0 - for soft interpolation (bilinear).
+                                     * 1 - for bicubic interpolation (4 tap Catmull-Rom).
+                                     * 2 - for sharper Wiener interpolation (6 tap, similar to Lanczos).
+                                    This parameter controls the calculation of the first level only.
+                                    When pel = 4, bilinear interpolation is always used to compute the second level.
+        :param rfilter:             Hierarchical levels smoothing and reducing (halving) filter.\n
+                                     * 0 - Simple 4 pixels averaging.
+                                     * 1 - Triangle (shifted) for more smoothing (decrease aliasing).
+                                     * 2 - Triangle filter like Bilinear for even more smoothing.
+                                     * 3 - Quadratic filter for even more smoothing.
+                                     * 4 - Cubic filter like Bicubic(b=1, c=0) for even more smoothing.
+        :param prefilter:           Prefilter to use for motion estimation. Can be a prefiltered clip instead.
+                                    The ideal prefiltered clip will be one that has little to not
+                                    temporal instability or dynamic grain, but retains most of the detail.
+        :param pel_type:            Type of interpolation to use for upscaling the pel clip.
+        :param ref:                 Reference clip to use for creating super clips.
+
+        :return:                    SuperClips yuple containing the render, search, and recalculate super clips.
+        """
+
+        ref = self.get_ref_clip(ref, self.__class__.super)
+        rfilter = kwargs_fallback(rfilter, (self.super_func_kwargs, 'rfilter'), 3)
+        range_conversion = kwargs_fallback(range_conversion, (self.super_func_kwargs, 'range_conversion'), 5.0)
+
+        sharp = kwargs_fallback(sharp, (self.super_func_kwargs, 'sharp'), 2)
+
+        prefilter = kwargs_fallback(  # type: ignore[assignment]
+            prefilter, (self.super_func_kwargs, 'prefilter'), Prefilter.AUTO
+        )
+
+        pel_type = kwargs_fallback(  # type: ignore[assignment]
+            pel_type, (self.super_func_kwargs, 'pel_type'), PelType.AUTO
+        )
+
+        if not isinstance(pel_type, tuple):
+            pel_type = (pel_type, pel_type)  # type: ignore[assignment]
+
+        if isinstance(prefilter, Prefilter):
+            prefilter = prefilter(ref, self.planes)
+
+            if self.range_in.is_limited:
+                prefilter = prefilter_to_full_range(prefilter, range_conversion, self.planes)
+
+        assert prefilter is not None
+
+        if self.high_precision:
+            prefilter = depth(prefilter, 32)
+
+        check_ref_clip(ref, prefilter)
+        pelclip, pelclip2 = self.get_subpel_clips(prefilter, ref, pel_type)  # type: ignore[arg-type]
+
+        common_args = dict[str, Any](
+            sharp=sharp, pel=self.pel, vpad=self.vpad_half, hpad=self.hpad_uhd, chroma=self.chroma
+        ) | self.super_args
+        super_render_args = common_args | dict(levels=1, hpad=self.hpad, vpad=self.vpad, chroma=not self.is_gray)
+
+        if pelclip or pelclip2:
+            common_args |= dict(pelclip=pelclip)
+            super_render_args |= dict(pelclip=pelclip2)
+
+        super_render = self.mvtools.Super(self.workclip, **super_render_args)
+        super_search = self.mvtools.Super(ref, **(dict(rfilter=rfilter) | common_args))
+        super_recalc = self.refine and self.mvtools.Super(prefilter, **(dict(levels=1) | common_args)) or super_render
+
+        supers = SuperClips(ref, super_render, super_search, super_recalc)
+
+        if not inplace:
+            self.supers = supers
+
+        return supers
+
+    def analyze(
+        self, block_size: int | None = None, overlap: int | None = None, thSAD: int | None = None,
+        search: SearchMode | SearchMode.Config | None = None,
+        sad_mode: SADMode | tuple[SADMode, SADMode] | None = None,
+        motion: MotionMode.Config | None = None, supers: SuperClips | None = None,
+        *, ref: vs.VideoNode | None = None, inplace: bool = False
     ) -> MotionVectors:
         """
         During the analysis stage, the plugin divides frames by small blocks and for every block in current frame
@@ -930,32 +1031,11 @@ class MVTools:
         :param thSAD:               During the recalculation, only bad quality new vectors with SAD above this thSAD
                                     will be re-estimated by search. thSAD value is scaled to 8x8 block size.
                                     Good vectors are not changed, but their SAD will be re-calculated and updated.
-        :param range_conversion:    If the input is limited, it will be converted to full range
-                                    to allow the motion analysis to use a wider array of information.\n
-                                    This is for deciding what range conversion method to use.
-                                     * >= 1.0 - Expansion with expr based on this coefficient.
-                                     * >  0.0 - Expansion with retinex.
-                                     * <= 0.0 - Simple conversion with resize plugin.
         :param search:              Decides the type of search at every level of the hierarchial
                                     analysis made while searching for motion vectors.
-        :param sharp:               Subpixel interpolation method for pel = 2 or 4. Possible values are 0, 1, 2.\n
-                                     * 0 - for soft interpolation (bilinear).
-                                     * 1 - for bicubic interpolation (4 tap Catmull-Rom).
-                                     * 2 - for sharper Wiener interpolation (6 tap, similar to Lanczos).
-                                    This parameter controls the calculation of the first level only.
-                                    When pel = 4, bilinear interpolation is always used to compute the second level.
-        :param rfilter:             Hierarchical levels smoothing and reducing (halving) filter.\n
-                                     * 0 - Simple 4 pixels averaging.
-                                     * 1 - Triangle (shifted) for more smoothing (decrease aliasing).
-                                     * 2 - Triangle filter like Bilinear for even more smoothing.
-                                     * 3 - Quadratic filter for even more smoothing.
-                                     * 4 - Cubic filter like Bicubic(b=1, c=0) for even more smoothing.
         :param sad_mode:            SAD Calculation mode.
         :param motion:              A preset or custom parameters values for truemotion/motion analysis mode.
-        :param prefilter:           Prefilter to use for motion estimation. Can be a prefiltered clip instead.
-                                    The ideal prefiltered clip will be one that has little to not
-                                    temporal instability or dynamic grain, but retains most of the detail.
-        :param pel_type:            Type of interpolation to use for upscaling the pel clip.
+        :param supers:              Custom super clips to be used for analyze.
         :param ref:                 Reference clip to use for analyzes over the main clip.
         :param inplace:             Whether to save the analysis in the MVTools instance or not.
 
@@ -972,14 +1052,8 @@ class MVTools:
 
         overlap = kwargs_fallback(overlap, (self.analyze_func_kwargs, 'overlap'), halfblocksize)
 
-        rfilter = kwargs_fallback(rfilter, (self.analyze_func_kwargs, 'rfilter'), 3)
-
         thSAD = kwargs_fallback(thSAD, (self.analyze_func_kwargs, 'thSAD'), 300)
 
-        range_conversion = kwargs_fallback(range_conversion, (self.analyze_func_kwargs, 'range_conversion'), 5.0)
-
-        sharp = kwargs_fallback(sharp, (self.analyze_func_kwargs, 'sharp'), 2)
-
         search = kwargs_fallback(  # type: ignore[assignment]
             search, (self.analyze_func_kwargs, 'search'),
             SearchMode.HEXAGON if self.refine else SearchMode.DIAMOND
@@ -1001,17 +1075,6 @@ class MVTools:
             sad_mode, (self.analyze_func_kwargs, 'sad_mode'), SADMode.SATD
         )
 
-        prefilter = kwargs_fallback(  # type: ignore[assignment]
-            prefilter, (self.analyze_func_kwargs, 'prefilter'), Prefilter.AUTO
-        )
-
-        pel_type = kwargs_fallback(  # type: ignore[assignment]
-            pel_type, (self.analyze_func_kwargs, 'pel_type'), PelType.AUTO
-        )
-
-        if not isinstance(pel_type, tuple):
-            pel_type = (pel_type, pel_type)  # type: ignore[assignment]
-
         vectors = MotionVectors() if inplace else self.vectors
 
         if isinstance(sad_mode, tuple):
@@ -1019,35 +1082,7 @@ class MVTools:
         else:
             sad_mode, recalc_sad_mode = sad_mode, SADMode.SATD
 
-        if isinstance(prefilter, Prefilter):
-            prefilter = prefilter(ref, self.planes)
-
-            if self.range_in.is_limited:
-                prefilter = prefilter_to_full_range(prefilter, range_conversion, self.planes)
-
-        assert prefilter is not None
-
-        if self.high_precision:
-            prefilter = depth(prefilter, 32)
-
-        check_ref_clip(ref, prefilter)
-
-        pelclip, pelclip2 = self.get_subpel_clips(prefilter, ref, pel_type)  # type: ignore[arg-type]
-
-        common_args = dict[str, Any](
-            sharp=sharp, pel=self.pel, vpad=self.vpad_half, hpad=self.hpad_uhd, chroma=self.chroma
-        ) | self.super_args
-        super_render_args = common_args | dict(levels=1, hpad=self.hpad, vpad=self.vpad, chroma=not self.is_gray)
-
-        if pelclip or pelclip2:
-            common_args |= dict(pelclip=pelclip)
-            super_render_args |= dict(pelclip=pelclip2)
-
-        super_search = self.mvtools.Super(ref, **(dict(rfilter=rfilter) | common_args))
-        super_render = self.mvtools.Super(self.workclip, **super_render_args)
-        super_recalc = self.mvtools.Super(
-            prefilter, **(dict(levels=1) | common_args)
-        ) if self.refine else super_render
+        supers = supers or self.get_supers(ref)
 
         if self.params_curve:
             thSAD_recalc = round(exp(-101. / (thSAD * 0.83)) * 360)
@@ -1059,30 +1094,36 @@ class MVTools:
         analyze_args = dict[str, Any](
             dct=sad_mode, pelsearch=search.pel, blksize=blocksize, overlap=overlap, search=search.mode,
             truemotion=motion.truemotion, searchparam=search.param, chroma=self.chroma,
-            plevel=motion.plevel, pglobal=motion.pglobal
+            plevel=motion.plevel, pglobal=motion.pglobal, pnew=motion.pnew,
+            lambda_=motion.block_coherence(blocksize), lsad=motion.sad_limit,
         ) | self.analyze_args
 
         recalc_args = dict[str, Any](
             search=search.recalc_mode, dct=recalc_sad_mode, thsad=thSAD_recalc, blksize=halfblocksize,
             overlap=halfoverlap, truemotion=motion.truemotion, searchparam=search.param_recalc,
-            chroma=self.chroma
+            chroma=self.chroma, pnew=motion.pnew, lambda_=motion.block_coherence(halfblocksize)
         ) | self.recalculate_args
 
         if self.mvtools is MVToolsPlugin.FLOAT_NEW:
-            vmulti = self.mvtools.Analyse(super_search, radius=t2, **analyze_args)
+            vmulti = self.mvtools.Analyse(supers.search, radius=t2, **analyze_args)
 
             vectors.vmulti = vmulti
 
             for i in range(self.refine):
-                recalc_args.update(blksize=blocksize / 2 ** i, overlap=blocksize / 2 ** (i + 1))
-                vectors.vmulti = self.mvtools.Recalculate(super_recalc, vectors.vmulti, **recalc_args)
+                val = clamp(blocksize / 2 ** i, 4, 128)
+                recalc_args.update(blksize=val, overlap=val / 2, lambda_=motion.block_coherence(val))
+                vectors.vmulti = self.mvtools.Recalculate(supers.recalculate, vectors.vmulti, **recalc_args)
         else:
             def _add_vector(delta: int, analyze: bool = True) -> None:
+                assert supers
+
                 for direction in MVDirection:
                     if analyze:
-                        vect = self.mvtools.Analyse(super_search, isb=direction.isb, delta=delta, **analyze_args)
+                        vect = self.mvtools.Analyse(supers.search, isb=direction.isb, delta=delta, **analyze_args)
                     else:
-                        vect = self.mvtools.Recalculate(super_recalc, vectors.get_mv(direction, delta), **recalc_args)
+                        vect = self.mvtools.Recalculate(
+                            supers.recalculate, vectors.get_mv(direction, delta), **recalc_args
+                        )
 
                     vectors.set_mv(direction, delta, vect)
 
@@ -1095,17 +1136,40 @@ class MVTools:
                         continue
 
                     for j in range(0, self.refine):
-                        val = clamp(blocksize / 2 ** j, 4, 128)
+                        recalc_blksize = clamp(blocksize / 2 ** j, 4, 128)
 
-                        recalc_args.update(blksize=val, overlap=val / 2)
+                        recalc_args.update(
+                            blksize=recalc_blksize, overlap=recalc_blksize // 2, lambda_=motion.block_coherence(val)
+                        )
 
                         _add_vector(i, False)
 
-        vectors.super_render = super_render
         vectors.kwargs.update(thSAD=thSAD)
 
         return vectors
 
+    def get_supers(self, ref: vs.VideoNode, *, inplace: bool = False) -> SuperClips:
+        """
+        Get the super clips for the specified ref clip.
+
+        If :py:attr:`analyze` wasn't previously called,
+        it will do so here with default values or kwargs specified in the constructor.
+
+        :param inplace:     Only return the SuperClips object, not modifying the internal state.
+
+        :return:            SuperClips tuple.
+        """
+
+        if self.supers and self.supers.base == ref:
+            return self.supers
+
+        supers = self.super(ref=ref)
+
+        if not inplace:
+            self.supers = supers
+
+        return supers
+
     def get_vectors_bf(self, *, inplace: bool = False) -> tuple[list[vs.VideoNode], list[vs.VideoNode]]:
         """
         Get the backwards and forward vectors.\n
@@ -1142,7 +1206,8 @@ class MVTools:
         return (vectors_backward, vectors_forward)
 
     def compensate(
-        self, func: GenericVSFunction, thSAD: int = 150, *, ref: vs.VideoNode | None = None, **kwargs: Any
+        self, func: GenericVSFunction, thSAD: int = 150,
+        supers: SuperClips | None = None, *, ref: vs.VideoNode | None = None, **kwargs: Any
     ) -> vs.VideoNode:
         """
         At compensation stage, the plugin client functions read the motion vectors and use them to move blocks
@@ -1161,6 +1226,7 @@ class MVTools:
         :param thSAD:       This is the SAD threshold for safe (dummy) compensation.\n
                             If block SAD is above thSAD, the block is bad, and we use source block
                             instead of the compensated block.
+        :param supers:      Custom super clips to be used for compensating.
         :param ref:         Reference clip to use instead of main clip.
         :param kwargs:      Keyword arguments passed to `func` to avoid using `partial`.
 
@@ -1170,9 +1236,10 @@ class MVTools:
         ref = self.get_ref_clip(ref, self.__class__.compensate)
 
         vect_b, vect_f = self.get_vectors_bf()
+        supers = supers or self.get_supers(ref)
 
         compensate_args = dict(
-            super=self.vectors.super_render, thsad=thSAD,
+            super=supers.render, thsad=thSAD,
             tff=self.source_type.is_inter and self.source_type.value or None
         ) | self.compensate_args
 
@@ -1195,6 +1262,7 @@ class MVTools:
         thSAD: int | tuple[int | None, int | None] | None = None,
         limit: int | tuple[int, int] = 255,
         thSCD: int | tuple[int | None, int | None] | None = (None, 51),
+        supers: SuperClips | None = None,
         *, ref: vs.VideoNode | None = None
     ) -> vs.VideoNode:
         """
@@ -1232,6 +1300,7 @@ class MVTools:
                         the frame to be considered as a scene change. It ranges from 0 to 100 %.
         :param ref:     Reference clip to use rather than the main clip. If passed,
                         the degraining will be applied to the ref clip rather than the original input clip.
+        :param supers:  Custom super clips to be used for degraining.
 
         :return:        Degrained clip.
         """
@@ -1239,6 +1308,7 @@ class MVTools:
         ref = self.get_ref_clip(ref, self.__class__.degrain)
 
         vect_b, vect_f = self.get_vectors_bf()
+        supers = supers or self.get_supers(ref)
 
         thSAD, thSADC = (thSAD if isinstance(thSAD, tuple) else (thSAD, None))
 
@@ -1277,10 +1347,10 @@ class MVTools:
                 degrain_args.update(thsad2=[thSAD / 2, thSADC / 2])
 
         if self.mvtools is MVToolsPlugin.FLOAT_NEW:
-            output = self.mvtools.Degrain()(ref, self.vectors.super_render, self.vectors.vmulti, **degrain_args)
+            output = self.mvtools.Degrain()(ref, supers.render, self.vectors.vmulti, **degrain_args)
         else:
             output = self.mvtools.Degrain(self.tr)(
-                ref, self.vectors.super_render, *chain.from_iterable(zip(vect_b, vect_f)), **degrain_args
+                ref, supers.render, *chain.from_iterable(zip(vect_b, vect_f)), **degrain_args
             )
 
         return output.std.DoubleWeave(self.source_type.value) if self.source_type.is_inter else output
@@ -1325,3 +1395,37 @@ class MVTools:
                 clip, self.pel, PelType.WIENER if is_ref else PelType.BICUBIC
             ) for is_ref, ptype, clip in zip((False, True), pel_type, (pref, ref))
         )
+
+    @classmethod
+    def denoise(
+        cls, clip: vs.VideoNode, thSAD: int | tuple[int, int | tuple[int, int]] | None = None,
+        tr: int = 2, refine: int = 3, block_size: int | None = None, overlap: int | None = None,
+        prefilter: Prefilter | vs.VideoNode | None = None, pel: int | None = None,
+        sad_mode: SADMode | tuple[SADMode, SADMode] | None = None,
+        search: SearchMode | SearchMode.Config | None = None, motion: MotionMode.Config | None = None,
+        pel_type: PelType | tuple[PelType, PelType] | None = None,
+        planes: PlanesT = None, range_in: ColorRange | None = None,
+        source_type: FieldBasedT | None = None, high_precision: bool = False,
+        limit: int | tuple[int, int] = 255, thSCD: int | tuple[int | None, int | None] | None = (None, 51),
+        *, super_args: dict[str, Any] | None = None, analyze_args: dict[str, Any] | None = None,
+        recalculate_args: dict[str, Any] | None = None, compensate_args: dict[str, Any] | None = None,
+        range_conversion: float | None = None, sharp: int | None = None,
+        hpad: int | None = None, vpad: int | None = None, params_curve: bool = True,
+        rfilter: int | None = None, vectors: MotionVectors | MVTools | None = None, supers: SuperClips | None = None
+    ) -> vs.VideoNode:
+        mvtools = cls(
+            clip, tr, refine, pel, planes, range_in, source_type, high_precision, hpad, vpad,
+            vectors, params_curve, super_args=super_args, analyze_args=analyze_args,
+            recalculate_args=recalculate_args, compensate_args=compensate_args
+        )
+
+        if isinstance(thSAD, Sequence):
+            thSADA, thSADD = thSAD  # type: ignore
+        else:
+            thSADA = thSADD = thSAD
+
+        supers = supers or mvtools.super(range_conversion, sharp, rfilter, prefilter, pel_type)
+
+        vectors = vectors or mvtools.analyze(block_size, overlap, thSADA, search, sad_mode, motion, supers)
+
+        return mvtools.degrain(thSADD, limit, thSCD, supers)
